
#include "definitions.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>

#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "uart.h"
#include "uexpect.h"


#define COUNTER_RED_LED OCR2B //Output Compare Register 2A (8 Bit) - Pin PD3
#define COUNTER_GREEN_LED OCR0B //Output Compare Register 0B (16 Bit, PWM aber 8 Bit) - Pin PD5
#define COUNTER_BLUE_LED OCR0A //Output Compare Register 0A (16 Bit, PWM aber 8 Bit) - Pin PD6


#define ARRAY_SIZE(array) (sizeof(array) / sizeof(*array))

/** Setting uart in- and output to stdin / stdout. so we can reuse existing functions like printf */
FILE uart_output = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE); 
FILE uart_input = FDEV_SETUP_STREAM(NULL, uart_getchar, _FDEV_SETUP_READ);
FILE uart_io = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

int currentMode = 0; 
const int MODE_SETRGB = 1;
const int MODE_FADE = 2; 
const char KEY_SETRGB[] = "setRGB";
const char KEY_FADE[] = "fade";
const char INITIAL_ACK[] = "ESP8266 UP AND RUNNING";
const char MSG_FINISHED[] = "ESP_FINISHED\r\n";
int isinitcomplete = 0; 

//global variables for color fader
int hue = 0; 
int saturation = 0; 
int intensity = 0; 

/**
* Pins auf Output setzen
* OC0A sitzt auf Pin 12 (PD6), OCOB auf Pin 11 (PD5) und OC2B auf Pin 5 (PD3) -- DDR = Data Direction Register, 0 = Input, 1 = Output
*/
void setPwmPinsEnabled(int isEnabled) {
	if (isEnabled == 1) {
		DDRD = 0 | (1 << DDD6) | (1 << DDD5) | (1 << DDD3);
	} else {
		DDRD  =  0;
	}
}

int clamp(int value, int maxValue) {
	if (value < 0) {
		return 0; 
	} else {
		if (value > maxValue) {
			return maxValue;
		}
	}
	
	return value;
}

void setRGB(int r, int g, int b) {
	if (r > 0 || g > 0 || b > 0) {
		setPwmPinsEnabled(1);
	} else {
		setPwmPinsEnabled(0);
	}
	
	COUNTER_RED_LED = clamp(r, 255);
	COUNTER_GREEN_LED = clamp(g, 255);
	COUNTER_BLUE_LED = clamp(b, 255);
}

void setR() {
	setRGB(255,0,0);
}

void setG() {
	setRGB(0, 255, 0);
}

void setB() {
	setRGB(0, 0, 255);
}

void blinkR(int times) {
	for(int i=0; i < times; i++) {
		setR();
		_delay_ms(200);
		setRGB(0,0,0);
		_delay_ms(200);
	}
}

void blinkG(int times) {
	for(int i=0; i < times; i++) {
		setG();
		_delay_ms(200);
		setRGB(0,0,0);
		_delay_ms(200);
	}
}

void blinkB(int times) {
	for(int i=0; i < times; i++) {
		setB();
		_delay_ms(200);
		setRGB(0,0,0);
		_delay_ms(200);
	}
}

/** 
 *  returns the data-block inside data={}
 *  DO NOT FORGET TO FREE THE RETURN VALUE
 */
char* extractData(char* haystack, char* needle) {
	char dataPart[strlen(needle) + 2];
	strcpy(dataPart, needle);
	strcat(dataPart, "={");
	int offset = (int)strlen(dataPart);
	char* begin = strstr(haystack, dataPart);
	
	if (begin != NULL) {
	    char* end = strstr(begin, "}");
		if (end != NULL) {
		    //length of data block inside {}
			int strlen = (int)(end - begin - offset);
			
			char *returnValue = malloc(strlen+1);
			if (!returnValue) {
			    return NULL; 
			}
			
			//null-terminate the string
			returnValue[strlen] = '\0';
			strncpy(returnValue, begin+offset, strlen);
			//memcpy(dataOnly, data, strlen+1);
		
			return returnValue;
		}
    }
	return NULL;
}

void handleFader(char* data) {
	char *hsiToken;
	
	while((hsiToken = strsep(&data, ";")) != NULL) {
		char param = hsiToken[0];
		if (param == 'h') {
			hue = clamp(atoi(hsiToken + 2), 360);
		} else if (param == 's') {
			saturation = clamp(atoi(hsiToken + 2), 100);
		} else if (param == 'i') {
			intensity = clamp(atoi(hsiToken + 2), 100);
		}
	}
	
	currentMode = MODE_FADE;
}

void handleSetRGB(char* data) {
	printf("rgb data: %s\r\n", data);
	char *rgbToken;
	int red = 0, green = 0, blue = 0, white = 0; 
	
	while((rgbToken = strsep(&data, ";")) != NULL) {
		printf("token: %s\r\n", rgbToken);
        char color = rgbToken[0];
		int value = atoi(rgbToken + 2);
		
		if (color == 'r') {
			red = value;
		} else if (color == 'g') {
		    green = value;
		} else if (color == 'b') {
	    	blue = value;
		} else if (color == 'w') {
		    white = value; 	   
		}

	}
	
	currentMode = MODE_SETRGB;
	setRGB(red, green, blue);

	//todo: erweitern auf wei, dazu vierten PWM-Ausgang machen 
}

void setupPwm() {
	//Pins auf Output setzen
	setPwmPinsEnabled(1);
	
	//Timer/Counter 0 Control Register (Register A und B) - 8 Bit each
	//Hier setzt man eigentlich nur die Settings, das hat nichts mit dem rauf/runterzaehlen zu tun. Daher reichen 8 Bit auch vllig aus
	
	/* TCCR0A = TC0 Control Register A */
	//Fr FAST PWM: COM0A1 = 1 und COM0A0 = 1: Set OC0A on compare match, clear OC0A at BOTTOM, (inverting mode)
	//D.h. je hoeher der Wert im TCCR, desto krzer leuchtet die LED
	TCCR0A = 0 | (1 << COM0A1) | (1 << COM0A0);
	
	//Selbiges gilt fuer COM0B1 und COM0B0
	TCCR0A |= (1 << COM0B1) | (1 << COM0B0);
	
	//WGM01 und WGM11 = 1 -> Fast PWM
	TCCR0A |= (1 << WGM01) | (1 << WGM00);
	
	/* TCCR0B TC0 Control Register B */
	//Nicht das Register fuer zweiten Counter, da das schon in TCCR0A mitkonfiguriert ist (COB0B1 und COM0B0).
	
	//FOC0A und FOC0B haben nur bei non-pwm Modes eine Funktion, mssen aber zwecks Kompatibilitt auf 0 gesetzt werden 
	TCCR0B = 0 | (0 << FOC0A) | (0 << FOC0B); 
	
	//WGM02 auf 0, somit ist TOP = 0xff, und der compare match triggert beim Wert von OCR0A bzw. OCR0B. Wenn WGM02 = 1, dann ist top = OCR0A und fr den compare match wird OCR0B herangezogen
	TCCR0B |= (0 << WGM02);
	
	/*
	*	CS12		CS11		CS10		Prescaler
	*	0			0			0			No clock source / timer stopped			
	*	0						1			No prescaling
	*	0			1			0			Prescaler 8
	*	0			1			1			Prescaler 64
		
		mehr dazu im Datasheet auf Seite 173
	*/
	
	TCCR0B |= (0 << CS02) | (0 << CS01) | (1 << CS00);

	/************************************************************************/

	/* Ab hier Timer/Counter 2                                              */

	/************************************************************************/
	
	/* Timer/Counter 2 Control Register A */
	//wenn Fast PWM: inverted mode
	TCCR2A = 0 | (1 << COM2B1) | (1 << COM2B0);
	
	//PWM auf drittem Channel (OC2A) bleibt erstmal deaktiviert, da nicht bentigt
	
	//Fast PWM
	TCCR2A |= (1 << WGM21) | (1 << WGM20);
	
	//Wieder die ungenutzten Flags auf 0 setzen 
	TCCR2B = 0 | (0 < FOC2A) | (0 < FOC2B);
	
	//WGM22 ebenfalls auf 0, damit Register f. compare match Verwendung findet
	TCCR2B |= (0 < WGM22); 
	
	//und ebenfalls kein Prescaler
	TCCR2B |= (0 < CS22) | (0 < CS21) | (1 < CS20);
}

void hsi2rgb(float H, float S, float I, int* rgb) {
	int r, g, b;
	H = 3.1415 * H / (float)180; // Convert to radians.
	S = S>0?(S<1?S:1):0; // clamp S and I to interval [0,1]
	I = I>0?(I<1?I:1):0;
	
	// Math! Thanks in part to Kyle Miller.
	if(H < 2.09439) {
		r = 255*I/3*(1+S*cos(H)/cos(1.047196667-H));
		g = 255*I/3*(1+S*(1-cos(H)/cos(1.047196667-H)));
		b = 255*I/3*(1-S);
	} else if(H < 4.188787) {
		H = H - 2.09439;
		g = 255*I/3*(1+S*cos(H)/cos(1.047196667-H));
		b = 255*I/3*(1+S*(1-cos(H)/cos(1.047196667-H)));
		r = 255*I/3*(1-S);
	} else {
		H = H - 4.188787;
		b = 255*I/3*(1+S*cos(H)/cos(1.047196667-H));
		r = 255*I/3*(1+S*(1-cos(H)/cos(1.047196667-H)));
		g = 255*I/3*(1-S);
	}
	
	rgb[0] = r;
	rgb[1] = g;
	rgb[2] = b;
}

ISR(BADISR_vect) {
	printf("unexpected vector catch-all triggered\r\n");
}
/*
 OC1A, OC1B und OC2 sind die Ausgabepins der PWM                      
 OCR1A etc sind die zugehrigen Compare Registers 
 Dabei wird der Counter inkrementiert und bei Match bit OCRn wird OCn gesetzt.                   
*/

int main(void) {
	
	setupPwm();
	int8_t returnValue;
	
	uart_init(0);
	stdin = &uart_input; 
	stdout = &uart_output;

	//Speicher f. RGB-Werte reservieren
	
	printf("Restarting ATMEGA328P\r\n");
	
	blinkR(1); 
	blinkG(1); 
	blinkB(1);
		
	_delay_ms(2000);
	
	while(1) {
		
		if(getReadBufferSize() > 0) {
			printf("message waiting (%d)\r\n", getReadBufferSize());
			//wait for initial acknowledgment that ESP01 is up and running to ignore all the spam fired when the chip is booting 
			returnValue = uexpect(stdin, MSG_FINISHED, NULL); 
			printf("message received: %s\r\n", uexpect_before);
			if (returnValue == -1) {
				blinkR(10);
				return -1;
			}
		
			if (isinitcomplete == 0) {
				//check if initial acknowledgment arrived and print to stdout
				char* result = strstr(uexpect_before, INITIAL_ACK);
				if (result != NULL) {
					isinitcomplete = 1; 
					blinkG(5);
					continue; 
				} else {
					//maybe the ATMEGA has rebooted and never gets an acknowledgment again? 
					isinitcomplete = 1;
					
				}
			}
			printf("reading data\r\n");
			
			char* path = extractData(uexpect_before, "path");
			char* dataPart = extractData(uexpect_before, "data");
			
			if(strcmp(KEY_SETRGB, path) == 0) {
				printf("Handle SETRGB\r\n");
				handleSetRGB(dataPart);
			} else if(strcmp(KEY_FADE, path) == 0) {
				printf("handleFader\r\n");
				handleFader(dataPart);
			} else {
				printf("unknown command %s\r\n", uexpect_before);
			}
			
			free(path);
			free(dataPart);
		
		} else {
			
			//buffer is empty, so stick to current mode (only if fading is enabled) 
			if (currentMode == MODE_FADE) {
				int *rgb = (int *)calloc(3, sizeof(int));
				hsi2rgb(hue, saturation, intensity, rgb);
				//printf("inside loop | h: %d, s: %d, i: %d, r: %d, g: %d, b: %d\r\n", hue, saturation, intensity, rgb[0], rgb[1], rgb[2]);
				setRGB(rgb[0], rgb[1], rgb[2]);
				free(rgb);
				hue = fmod(hue + 1, 360); // cycle H around to 0-360 degrees;
				
				//_delay_ms(100);
			
			} else {
				//printf("current mode: %d\r\n", currentMode);
			}
		}
	 }
	 
	 printf("terminating main");
	 
}


